@@ -1,6 +1,5 @@
 #include "pgroonga.h"
 
-#include "pgrn-compatible.h"
 #include "pgrn-global.h"
 #include "pgrn-groonga.h"
 #include "pgrn-jsonb.h"
@@ -18,6 +17,7 @@ typedef struct PGrnOptions
 {
 	int32 vl_len_;
 	int tokenizerOffset;
+	int tokenizerMappingOffset;
 	int normalizerOffset;
 	int tokenFiltersOffset;
 	int pluginsOffset;
@@ -29,20 +29,29 @@ typedef struct PGrnOptions
 	int normalizersOffset;
 	int normalizersMappingOffset;
 	int indexFlagsMappingOffset;
+	int modelOffset;
+	int nGPULayers;
+	int passagePrefixOffset;
+	int queryPrefixOffset;
 } PGrnOptions;
 
 static relopt_kind PGrnReloptionKind;
 
-static grn_ctx *ctx = &PGrnContext;
 static struct PGrnBuffers *buffers = &PGrnBuffers;
 static grn_obj *lexicon = NULL;
-
-PGDLLEXPORT PG_FUNCTION_INFO_V1(pgroonga_options);
+static grn_language_model_loader *languageModelLoader = NULL;
 
 typedef void (*PGrnOptionNameFunction)(const char *name,
 									   size_t nameSize,
 									   void *data);
 
+static void
+PGrnOptionAppendName(const char *name, size_t nameSize, void *data)
+{
+	grn_obj *vector = data;
+	grn_vector_add_element(ctx, vector, name, nameSize, 0, GRN_DB_TEXT);
+}
+
 static void
 PGrnOptionParseNames(const char *names,
 					 PGrnOptionNameFunction function,
@@ -70,7 +79,8 @@ PGrnOptionParseNames(const char *names,
 		}
 	}
 
-	if (current > start) {
+	if (current > start)
+	{
 		function(start, current - start, data);
 	}
 }
@@ -82,13 +92,13 @@ PGrnOptionEnsureLexicon(const char *context)
 		grn_obj_close(ctx, lexicon);
 
 	lexicon = grn_table_create(ctx,
-							   NULL, 0,
+							   NULL,
+							   0,
 							   NULL,
 							   GRN_TABLE_PAT_KEY,
 							   grn_ctx_at(ctx, GRN_DB_SHORT_TEXT),
 							   NULL);
-	PGrnCheck("options: %s: failed to create a temporary lexicon",
-			  context);
+	PGrnCheck("options: %s: failed to create a temporary lexicon", context);
 }
 
 static void
@@ -106,13 +116,76 @@ PGrnOptionValidateTokenizer(const char *rawTokenizer)
 	PGrnOptionEnsureLexicon("tokenizer");
 
 	GRN_TEXT_SETS(ctx, tokenizer, rawTokenizer);
-	grn_obj_set_info(ctx,
-					 lexicon,
-					 GRN_INFO_DEFAULT_TOKENIZER,
-					 tokenizer);
-	PGrnCheck("%s invalid tokenizer: <%s>",
-			  tag,
-			  rawTokenizer);
+	grn_obj_set_info(ctx, lexicon, GRN_INFO_DEFAULT_TOKENIZER, tokenizer);
+	PGrnCheck("%s invalid tokenizer: <%s>", tag, rawTokenizer);
+}
+
+static void
+PGrnOptionValidateTokenizerMapping(const char *rawTokenizerMapping)
+{
+	const char *tag = "[option][tokenizer-mapping][validate]";
+	grn_obj *tokenizer = &(buffers->tokenizer);
+	Jsonb *jsonb;
+	JsonbIterator *iter;
+	JsonbIteratorToken token;
+	JsonbValue value;
+
+	if (PGrnIsNoneValue(rawTokenizerMapping))
+		return;
+
+	jsonb = PGrnJSONBParse(rawTokenizerMapping);
+	iter = JsonbIteratorInit(&(jsonb->root));
+
+	PGrnOptionEnsureLexicon("tokenizer");
+
+	token = JsonbIteratorNext(&iter, &value, false);
+	if (token != WJB_BEGIN_OBJECT)
+	{
+		PGrnCheckRC(GRN_INVALID_ARGUMENT,
+					"%s must be object: %s: <%s>",
+					tag,
+					PGrnJSONBIteratorTokenToString(token),
+					rawTokenizerMapping);
+	}
+
+	while (true)
+	{
+		token = JsonbIteratorNext(&iter, &value, false);
+		if (token == WJB_END_OBJECT)
+			break;
+		if (token != WJB_KEY)
+		{
+			PGrnCheckRC(GRN_INVALID_ARGUMENT,
+						"%s misses key: %s: <%s>",
+						tag,
+						PGrnJSONBIteratorTokenToString(token),
+						rawTokenizerMapping);
+		}
+		token = JsonbIteratorNext(&iter, &value, false);
+		if (token != WJB_VALUE)
+		{
+			PGrnCheckRC(GRN_INVALID_ARGUMENT,
+						"%s misses value: %s: <%s>",
+						tag,
+						PGrnJSONBIteratorTokenToString(token),
+						rawTokenizerMapping);
+		}
+		if (value.type != jbvString)
+		{
+			PGrnCheckRC(GRN_INVALID_ARGUMENT,
+						"%s value must be string: %s: <%s>",
+						tag,
+						PGrnJSONBValueTypeToString(value.type),
+						rawTokenizerMapping);
+		}
+		GRN_TEXT_SET(
+			ctx, tokenizer, value.val.string.val, value.val.string.len);
+		grn_obj_set_info(ctx, lexicon, GRN_INFO_DEFAULT_TOKENIZER, tokenizer);
+		PGrnCheck("%s value is invalid tokenizer: <%.*s>",
+				  tag,
+				  (int) GRN_TEXT_LEN(tokenizer),
+				  GRN_TEXT_VALUE(tokenizer));
+	}
 }
 
 static void
@@ -130,13 +203,8 @@ PGrnOptionValidateNormalizers(const char *rawNormalizers)
 	PGrnOptionEnsureLexicon("normalizers");
 
 	GRN_TEXT_SETS(ctx, normalizers, rawNormalizers);
-	grn_obj_set_info(ctx,
-					 lexicon,
-					 GRN_INFO_NORMALIZERS,
-					 normalizers);
-	PGrnCheck("%s invalid normalizers: <%s>",
-			  tag,
-			  rawNormalizers);
+	grn_obj_set_info(ctx, lexicon, GRN_INFO_NORMALIZERS, normalizers);
+	PGrnCheck("%s invalid normalizers: <%s>", tag, rawNormalizers);
 }
 
 static void
@@ -167,7 +235,8 @@ PGrnOptionValidateNormalizersMapping(const char *rawNormalizersMapping)
 					rawNormalizersMapping);
 	}
 
-	while (true) {
+	while (true)
+	{
 		token = JsonbIteratorNext(&iter, &value, false);
 		if (token == WJB_END_OBJECT)
 			break;
@@ -196,14 +265,9 @@ PGrnOptionValidateNormalizersMapping(const char *rawNormalizersMapping)
 						PGrnJSONBValueTypeToString(value.type),
 						rawNormalizersMapping);
 		}
-		GRN_TEXT_SET(ctx,
-					 normalizers,
-					 value.val.string.val,
-					 value.val.string.len);
-		grn_obj_set_info(ctx,
-						 lexicon,
-						 GRN_INFO_NORMALIZERS,
-						 normalizers);
+		GRN_TEXT_SET(
+			ctx, normalizers, value.val.string.val, value.val.string.len);
+		grn_obj_set_info(ctx, lexicon, GRN_INFO_NORMALIZERS, normalizers);
 		PGrnCheck("%s value is invalid normalizer: <%.*s>",
 				  tag,
 				  (int) GRN_TEXT_LEN(normalizers),
@@ -223,37 +287,21 @@ PGrnOptionValidateTokenFilters(const char *rawTokenFilters)
 	PGrnOptionEnsureLexicon("token filters");
 
 	GRN_TEXT_SETS(ctx, tokenFilters, rawTokenFilters);
-	grn_obj_set_info(ctx,
-						  lexicon,
-						  GRN_INFO_TOKEN_FILTERS,
-						  tokenFilters);
-	PGrnCheck("%s invalid token filters: <%s>",
-			  tag,
-			  rawTokenFilters);
+	grn_obj_set_info(ctx, lexicon, GRN_INFO_TOKEN_FILTERS, tokenFilters);
+	PGrnCheck("%s invalid token filters: <%s>", tag, rawTokenFilters);
 }
 
 static void
-PGrnOptionValidatePlugin(const char *name,
-						 size_t nameSize,
-						 void *data)
+PGrnOptionValidatePlugin(const char *name, size_t nameSize, void *data)
 {
 	const char *tag = "[option][plugin][validate]";
-	char pluginName[MAXPGPATH];
-
-	grn_strncpy(pluginName, MAXPGPATH, name, nameSize);
-	pluginName[nameSize] = '\0';
-	grn_plugin_register(ctx, pluginName);
-	PGrnCheck("%s failed to register plugin: <%.*s>",
-			  tag,
-			  (int) nameSize, name);
+	PGrnRegisterPluginWithSize(name, nameSize, tag);
 }
 
 static void
 PGrnOptionValidatePlugins(const char *names)
 {
-	PGrnOptionParseNames(names,
-						 PGrnOptionValidatePlugin,
-						 NULL);
+	PGrnOptionParseNames(names, PGrnOptionValidatePlugin, NULL);
 }
 
 static void
@@ -309,7 +357,8 @@ PGrnOptionValidateIndexFlagsMapping(const char *rawIndexFlagsMapping)
 					rawIndexFlagsMapping);
 	}
 
-	while (true) {
+	while (true)
+	{
 		token = JsonbIteratorNext(&iter, &value, false);
 		if (token == WJB_END_OBJECT)
 			break;
@@ -364,98 +413,167 @@ PGrnOptionValidateIndexFlagsMapping(const char *rawIndexFlagsMapping)
 	}
 }
 
+static void
+PGrnOptionValidateModel(const char *rawModel)
+{
+	const char *tag = "[option][model][validate]";
+	grn_language_model *model;
+
+	if (!rawModel || rawModel[0] == '\0')
+		return;
+
+	if (!PGrnIsLlamaCppAvailable)
+		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
+					"%s Groonga's llama.cpp support isn't enabled",
+					tag);
+
+	if (!PGrnIsFaissAvailable)
+		PGrnCheckRC(GRN_FUNCTION_NOT_IMPLEMENTED,
+					"%s Groonga's Faiss support isn't enabled",
+					tag);
+
+	grn_language_model_loader_set_model(
+		ctx, languageModelLoader, rawModel, strlen(rawModel));
+#if GRN_VERSION_OR_LATER(15, 2, 1)
+	grn_language_model_loader_set_n_gpu_layers(ctx, languageModelLoader, 0);
+#endif
+	model = grn_language_model_loader_load(ctx, languageModelLoader);
+	if (model)
+	{
+		grn_language_model_close(ctx, model);
+	}
+	else
+	{
+		PGrnCheck("%s failed to load model: <%s>", tag, rawModel);
+	}
+}
+
 void
 PGrnInitializeOptions(void)
 {
-#ifdef PGRN_SUPPORT_OPTION_LOCK_MODE
 	const LOCKMODE lock_mode = ShareUpdateExclusiveLock;
-#endif
 
 	lexicon = NULL;
+	languageModelLoader = grn_language_model_loader_open(ctx);
 	PGrnReloptionKind = add_reloption_kind();
 
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "tokenizer",
-							  "Tokenizer name to be used for full-text search",
-							  PGRN_DEFAULT_TOKENIZER,
-							  PGrnOptionValidateTokenizer,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "normalizer",
-							  "Normalizers to be used as fallback. "
-							  "This is deprecated since 2.3.1. "
-							  "Use normalizers instead",
-							  NULL,
-							  PGrnOptionValidateNormalizers,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "token_filters",
-							  "Token filter names separated by \",\" "
-							  "to be used for full-text search",
-							  "",
-							  PGrnOptionValidateTokenFilters,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "plugins",
-							  "Plugin names separated by \",\" to be installed",
-							  "",
-							  PGrnOptionValidatePlugins,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "full_text_search_normalizer",
-							  "Normalizers to be used for full-text search. "
-							  "This is deprecated since 2.3.1. "
-							  "Use normalizers_mapping instead",
-							  NULL,
-							  PGrnOptionValidateNormalizers,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "regexp_search_normalizer",
-							  "Normalizers to be used for regexp search. "
-							  "This is deprecated since 2.3.1. "
-							  "Use normalizers_mapping instead",
-							  NULL,
-							  PGrnOptionValidateNormalizers,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "prefix_search_normalizer",
-							  "Normalizers to be used for prefix search"
-							  "This is deprecated since 2.3.1. "
-							  "Use normalizers_mapping instead",
-							  NULL,
-							  PGrnOptionValidateNormalizers,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "lexicon_type",
-							  "Lexicon type to be used for lexicon",
-							  NULL,
-							  PGrnOptionValidateLexiconType,
-							  lock_mode);
-	pgrn_add_bool_reloption(PGrnReloptionKind,
-							"query_allow_column",
-							"Accept column:... syntax in query",
-							false,
-							lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "normalizers",
-							  "Normalizers to be used as fallback",
-							  NULL,
-							  PGrnOptionValidateNormalizers,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "normalizers_mapping",
-							  "Mapping to specify normalizers to be used "
-							  "for each target",
-							  NULL,
-							  PGrnOptionValidateNormalizersMapping,
-							  lock_mode);
-	pgrn_add_string_reloption(PGrnReloptionKind,
-							  "index_flags_mapping",
-							  "Mapping to specify index flags to be used "
-							  "for each target",
-							  NULL,
-							  PGrnOptionValidateIndexFlagsMapping,
-							  lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "tokenizer",
+						 "Tokenizer name to be used for full-text search",
+						 PGRN_DEFAULT_TOKENIZER,
+						 PGrnOptionValidateTokenizer,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "tokenizer_mapping",
+						 "Mapping to specify tokenizer to be used "
+						 "for each target",
+						 NULL,
+						 PGrnOptionValidateTokenizerMapping,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "normalizer",
+						 "Normalizers to be used as fallback. "
+						 "This is deprecated since 2.3.1. "
+						 "Use normalizers instead",
+						 NULL,
+						 PGrnOptionValidateNormalizers,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "token_filters",
+						 "Token filter names separated by \",\" "
+						 "to be used for full-text search",
+						 "",
+						 PGrnOptionValidateTokenFilters,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "plugins",
+						 "Plugin names separated by \",\" to be installed",
+						 "",
+						 PGrnOptionValidatePlugins,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "full_text_search_normalizer",
+						 "Normalizers to be used for full-text search. "
+						 "This is deprecated since 2.3.1. "
+						 "Use normalizers_mapping instead",
+						 NULL,
+						 PGrnOptionValidateNormalizers,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "regexp_search_normalizer",
+						 "Normalizers to be used for regexp search. "
+						 "This is deprecated since 2.3.1. "
+						 "Use normalizers_mapping instead",
+						 NULL,
+						 PGrnOptionValidateNormalizers,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "prefix_search_normalizer",
+						 "Normalizers to be used for prefix search"
+						 "This is deprecated since 2.3.1. "
+						 "Use normalizers_mapping instead",
+						 NULL,
+						 PGrnOptionValidateNormalizers,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "lexicon_type",
+						 "Lexicon type to be used for lexicon",
+						 NULL,
+						 PGrnOptionValidateLexiconType,
+						 lock_mode);
+	add_bool_reloption(PGrnReloptionKind,
+					   "query_allow_column",
+					   "Accept column:... syntax in query",
+					   false,
+					   lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "normalizers",
+						 "Normalizers to be used as fallback",
+						 NULL,
+						 PGrnOptionValidateNormalizers,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "normalizers_mapping",
+						 "Mapping to specify normalizers to be used "
+						 "for each target",
+						 NULL,
+						 PGrnOptionValidateNormalizersMapping,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "index_flags_mapping",
+						 "Mapping to specify index flags to be used "
+						 "for each target",
+						 NULL,
+						 PGrnOptionValidateIndexFlagsMapping,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "model",
+						 "Language model name to be used by semantic search",
+						 NULL,
+						 PGrnOptionValidateModel,
+						 lock_mode);
+#ifndef GRN_LANGUAGE_MODEL_LOADER_N_GPU_LAYERS_DEFAULT
+#	define GRN_LANGUAGE_MODEL_LOADER_N_GPU_LAYERS_DEFAULT 999
+#endif
+	add_int_reloption(PGrnReloptionKind,
+					  "n_gpu_layers",
+					  "The number of layers used by language model",
+					  GRN_LANGUAGE_MODEL_LOADER_N_GPU_LAYERS_DEFAULT,
+					  0,
+					  GRN_LANGUAGE_MODEL_LOADER_N_GPU_LAYERS_DEFAULT,
+					  lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "passage_prefix",
+						 "Specify prefix for search target text",
+						 NULL,
+						 false,
+						 lock_mode);
+	add_string_reloption(PGrnReloptionKind,
+						 "query_prefix",
+						 "Specify prefix for query text",
+						 NULL,
+						 false,
+						 lock_mode);
 }
 
 void
@@ -464,16 +582,180 @@ PGrnFinalizeOptions(void)
 	if (lexicon)
 	{
 		grn_obj_close(ctx, lexicon);
+		lexicon = NULL;
+	}
+	if (languageModelLoader)
+	{
+		grn_language_model_loader_close(ctx, languageModelLoader);
+		languageModelLoader = NULL;
 	}
 }
 
 static void
-PGrnApplyOptionValuesNormalizers(PGrnOptions *options,
+PGrnResolveOptionValuesTokenizer(PGrnOptions *options,
 								 Relation index,
 								 int i,
 								 PGrnOptionUseCase useCase,
-								 grn_obj **normalizers,
-								 const char *defaultNormalizers)
+								 const char *defaultTokenizer,
+								 PGrnResolvedOptions *resolvedOptions)
+{
+	const char *rawTokenizer;
+	TupleDesc desc = RelationGetDescr(index);
+	Name name = &(TupleDescAttr(desc, i)->attname);
+
+	if (options->tokenizerMappingOffset != 0 && i >= 0)
+	{
+		const char *rawTokenizerMapping =
+			((const char *) options) + options->tokenizerMappingOffset;
+		Jsonb *jsonb = PGrnJSONBParse(rawTokenizerMapping);
+		JsonbIterator *iter;
+		JsonbValue value;
+
+		iter = JsonbIteratorInit(&(jsonb->root));
+		/* This JSON is validated. */
+		/* WJB_BEGIN_OBJECT */
+		JsonbIteratorNext(&iter, &value, false);
+		while (true)
+		{
+			bool isTarget;
+
+			/* WJB_KEY */
+			if (JsonbIteratorNext(&iter, &value, false) == WJB_END_OBJECT)
+				break;
+			isTarget = (value.val.string.len == strlen(name->data) &&
+						memcmp(value.val.string.val,
+							   name->data,
+							   value.val.string.len) == 0);
+			/* WJB_VALUE */
+			JsonbIteratorNext(&iter, &value, false);
+			if (!isTarget)
+				continue;
+
+			resolvedOptions->tokenizer = &(buffers->tokenizer);
+			GRN_BULK_REWIND(resolvedOptions->tokenizer);
+			PGrnStringSubstituteVariables(value.val.string.val,
+										  value.val.string.len,
+										  resolvedOptions->tokenizer);
+			return;
+		}
+	}
+
+	rawTokenizer = ((const char *) options) + options->tokenizerOffset;
+	if (useCase == PGRN_OPTION_USE_CASE_REGEXP_SEARCH)
+	{
+		resolvedOptions->tokenizer = PGrnLookup(defaultTokenizer, ERROR);
+	}
+	else if (useCase == PGRN_OPTION_USE_CASE_PREFIX_SEARCH)
+	{
+		resolvedOptions->tokenizer = NULL;
+	}
+	else if (useCase == PGRN_OPTION_USE_CASE_SEMANTIC_SEARCH)
+	{
+		char codeColumnName[GRN_TABLE_MAX_KEY_SIZE];
+		grn_language_model *model;
+
+		PGrnCodeColumnNameEncode(name->data, codeColumnName);
+
+		grn_language_model_loader_set_model(ctx,
+											languageModelLoader,
+											resolvedOptions->modelName,
+											strlen(resolvedOptions->modelName));
+#if GRN_VERSION_OR_LATER(15, 2, 1)
+		grn_language_model_loader_set_n_gpu_layers(
+			ctx, languageModelLoader, resolvedOptions->nGPULayers);
+#endif
+		model = grn_language_model_loader_load(ctx, languageModelLoader);
+		if (!model)
+		{
+			PGrnCheck("failed to load model: <%s>", resolvedOptions->modelName);
+		}
+#if GRN_VERSION_OR_LATER(15, 2, 1)
+		/* float * 1024 dimensions == 4KiB == The max key size
+		 *
+		 * If this model uses 1025 or more dimensions, we can't store
+		 * a centroid as a lexicon key. In the case, we need to store
+		 * a centroid as a column value.
+		 */
+		resolvedOptions->needCentroidColumn =
+			(grn_language_model_get_n_embedding_dimensions(ctx, model) > 1024);
+#else
+		resolvedOptions->needCentroidColumn = false;
+#endif
+
+		resolvedOptions->tokenizer = &(buffers->tokenizer);
+		GRN_BULK_REWIND(resolvedOptions->tokenizer);
+		grn_text_printf(ctx,
+						resolvedOptions->tokenizer,
+						"TokenLanguageModelKNN(\"model\", \"%s\", "
+						"\"code_column\", \"%s\", "
+						"\"n_gpu_layers\", %d",
+						resolvedOptions->modelName,
+						codeColumnName,
+						resolvedOptions->nGPULayers);
+
+		if (resolvedOptions->passagePrefix)
+		{
+			grn_text_printf(ctx,
+							resolvedOptions->tokenizer,
+							", \"passage_prefix\", \"%s\"",
+							resolvedOptions->passagePrefix);
+		}
+
+		if (resolvedOptions->queryPrefix)
+		{
+			grn_text_printf(ctx,
+							resolvedOptions->tokenizer,
+							", \"query_prefix\", \"%s\"",
+							resolvedOptions->queryPrefix);
+		}
+
+		if (resolvedOptions->needCentroidColumn)
+		{
+			resolvedOptions->lexiconKeyTypeID = GRN_DB_UINT32;
+			grn_text_printf(ctx,
+							resolvedOptions->tokenizer,
+							", \"centroid_column\", \"%s\"",
+							PGrnCentroidColumnName);
+		}
+		else
+		{
+#if GRN_VERSION_OR_LATER(15, 1, 8)
+			resolvedOptions->lexiconKeyTypeID = GRN_DB_SHORT_BINARY;
+#else
+			resolvedOptions->lexiconKeyTypeID = GRN_DB_SHORT_TEXT;
+#endif
+		}
+		GRN_TEXT_PUTS(ctx, resolvedOptions->tokenizer, ")");
+	}
+	else
+	{
+		if (PGrnIsExplicitNoneValue(rawTokenizer))
+		{
+			resolvedOptions->tokenizer = NULL;
+		}
+		else if (PGrnIsNoneValue(rawTokenizer))
+		{
+			if (defaultTokenizer)
+				resolvedOptions->tokenizer =
+					PGrnLookup(defaultTokenizer, ERROR);
+			else
+				resolvedOptions->tokenizer = NULL;
+		}
+		else
+		{
+			resolvedOptions->tokenizer = &(buffers->tokenizer);
+			GRN_TEXT_SETS(ctx, resolvedOptions->tokenizer, rawTokenizer);
+		}
+	}
+}
+
+static void
+PGrnResolveOptionValuesNormalizers(PGrnOptions *options,
+								   Relation index,
+								   int i,
+								   PGrnOptionUseCase useCase,
+								   const char *defaultNormalizers,
+								   PGrnResolvedOptions *resolvedOptions)
 {
 	const char *rawNormalizers;
 
@@ -507,11 +789,11 @@ PGrnApplyOptionValuesNormalizers(PGrnOptions *options,
 			if (!isTarget)
 				continue;
 
-			*normalizers = &(buffers->normalizers);
-			GRN_BULK_REWIND(*normalizers);
+			resolvedOptions->normalizers = &(buffers->normalizers);
+			GRN_BULK_REWIND(resolvedOptions->normalizers);
 			PGrnStringSubstituteVariables(value.val.string.val,
 										  value.val.string.len,
-										  *normalizers);
+										  resolvedOptions->normalizers);
 			return;
 		}
 	}
@@ -537,22 +819,22 @@ PGrnApplyOptionValuesNormalizers(PGrnOptions *options,
 	case PGRN_OPTION_USE_CASE_FULL_TEXT_SEARCH:
 		if (options->fullTextSearchNormalizerOffset != 0)
 		{
-			rawNormalizers =
-				((const char *) options) + options->fullTextSearchNormalizerOffset;
+			rawNormalizers = ((const char *) options) +
+							 options->fullTextSearchNormalizerOffset;
 		}
 		break;
 	case PGRN_OPTION_USE_CASE_REGEXP_SEARCH:
 		if (options->regexpSearchNormalizerOffset != 0)
 		{
-			rawNormalizers =
-				((const char *) options) + options->regexpSearchNormalizerOffset;
+			rawNormalizers = ((const char *) options) +
+							 options->regexpSearchNormalizerOffset;
 		}
 		break;
 	case PGRN_OPTION_USE_CASE_PREFIX_SEARCH:
 		if (options->prefixSearchNormalizerOffset != 0)
 		{
-			rawNormalizers =
-				((const char *) options) + options->prefixSearchNormalizerOffset;
+			rawNormalizers = ((const char *) options) +
+							 options->prefixSearchNormalizerOffset;
 		}
 		break;
 	default:
@@ -561,42 +843,39 @@ PGrnApplyOptionValuesNormalizers(PGrnOptions *options,
 
 	if (PGrnIsExplicitNoneValue(rawNormalizers))
 	{
-		*normalizers = NULL;
+		resolvedOptions->normalizers = NULL;
 	}
 	else if (PGrnIsNoneValue(rawNormalizers))
 	{
 		if (defaultNormalizers)
 		{
-			*normalizers = &(buffers->normalizers);
-			GRN_BULK_REWIND(*normalizers);
+			resolvedOptions->normalizers = &(buffers->normalizers);
+			GRN_BULK_REWIND(resolvedOptions->normalizers);
 			PGrnStringSubstituteVariables(defaultNormalizers,
 										  strlen(defaultNormalizers),
-										  *normalizers);
+										  resolvedOptions->normalizers);
 		}
 		else
 		{
-			*normalizers = NULL;
+			resolvedOptions->normalizers = NULL;
 		}
 	}
 	else
 	{
-		*normalizers = &(buffers->normalizers);
-		GRN_BULK_REWIND(*normalizers);
+		resolvedOptions->normalizers = &(buffers->normalizers);
+		GRN_BULK_REWIND(resolvedOptions->normalizers);
 		PGrnStringSubstituteVariables(rawNormalizers,
 									  strlen(rawNormalizers),
-									  *normalizers);
+									  resolvedOptions->normalizers);
 	}
 }
 
 static void
-PGrnApplyOptionValuesIndexFlags(PGrnOptions *options,
-								Relation index,
-								int i,
-								grn_column_flags *indexFlags)
+PGrnResolveOptionValuesIndexFlags(PGrnOptions *options,
+								  Relation index,
+								  int i,
+								  PGrnResolvedOptions *resolvedOptions)
 {
-	if (!indexFlags)
-		return;
-
 	if (options->indexFlagsMappingOffset != 0 && i >= 0)
 	{
 		TupleDesc desc = RelationGetDescr(index);
@@ -635,15 +914,16 @@ PGrnApplyOptionValuesIndexFlags(PGrnOptions *options,
 				rawFlag.value = value.val.string.val;
 				rawFlag.length = value.val.string.len;
 				if (GRN_RAW_STRING_EQUAL_CSTRING(rawFlag, "SMALL"))
-					*indexFlags |= GRN_OBJ_INDEX_SMALL;
+					resolvedOptions->indexFlags |= GRN_OBJ_INDEX_SMALL;
 				else if (GRN_RAW_STRING_EQUAL_CSTRING(rawFlag, "MEDIUM"))
-					*indexFlags |= GRN_OBJ_INDEX_MEDIUM;
+					resolvedOptions->indexFlags |= GRN_OBJ_INDEX_MEDIUM;
 				else if (GRN_RAW_STRING_EQUAL_CSTRING(rawFlag, "LARGE"))
-					*indexFlags |= GRN_OBJ_INDEX_LARGE;
+					resolvedOptions->indexFlags |= GRN_OBJ_INDEX_LARGE;
 				else if (GRN_RAW_STRING_EQUAL_CSTRING(rawFlag, "WITH_WEIGHT"))
-					*indexFlags |= GRN_OBJ_WITH_WEIGHT;
-				else if (GRN_RAW_STRING_EQUAL_CSTRING(rawFlag, "WEIGHT_FLOAT32"))
-					*indexFlags |= GRN_OBJ_WEIGHT_FLOAT32;
+					resolvedOptions->indexFlags |= GRN_OBJ_WITH_WEIGHT;
+				else if (GRN_RAW_STRING_EQUAL_CSTRING(rawFlag,
+													  "WEIGHT_FLOAT32"))
+					resolvedOptions->indexFlags |= GRN_OBJ_WEIGHT_FLOAT32;
 			}
 			if (isTarget)
 				return;
@@ -652,83 +932,68 @@ PGrnApplyOptionValuesIndexFlags(PGrnOptions *options,
 }
 
 void
-PGrnApplyOptionValues(Relation index,
-					  int i,
-					  PGrnOptionUseCase useCase,
-					  grn_obj **tokenizer,
-					  const char *defaultTokenizerName,
-					  grn_obj **normalizers,
-					  const char *defaultNormalizers,
-					  grn_obj **tokenFilters,
-					  grn_table_flags *lexiconType,
-					  grn_column_flags *indexFlags)
+PGrnResolveOptionValues(Relation index,
+						int i,
+						PGrnOptionUseCase useCase,
+						const char *defaultTokenizer,
+						const char *defaultNormalizers,
+						PGrnResolvedOptions *resolvedOptions)
 {
 	PGrnOptions *options;
-	const char *rawTokenizer;
 	const char *rawTokenFilters;
 	const char *lexiconTypeName;
 
 	options = (PGrnOptions *) (index->rd_options);
 	if (!options)
 	{
-		if (defaultTokenizerName)
-			*tokenizer = PGrnLookup(defaultTokenizerName, ERROR);
+		if (defaultTokenizer)
+			resolvedOptions->tokenizer = PGrnLookup(defaultTokenizer, ERROR);
 		else
-			*tokenizer = NULL;
+			resolvedOptions->tokenizer = NULL;
 
 		if (defaultNormalizers)
 		{
-			*normalizers = &(buffers->normalizers);
-			GRN_TEXT_SETS(ctx, *normalizers, defaultNormalizers);
+			resolvedOptions->normalizers = &(buffers->normalizers);
+			GRN_TEXT_SETS(
+				ctx, resolvedOptions->normalizers, defaultNormalizers);
 		}
 		else
 		{
-			*normalizers = NULL;
+			resolvedOptions->normalizers = NULL;
 		}
 
-		*lexiconType |= GRN_OBJ_TABLE_PAT_KEY;
+		resolvedOptions->lexiconType |= GRN_OBJ_TABLE_PAT_KEY;
 
 		return;
 	}
 
-	rawTokenizer = ((const char *) options) + options->tokenizerOffset;
+	resolvedOptions->modelName = GET_STRING_RELOPTION(options, modelOffset);
+	resolvedOptions->nGPULayers = options->nGPULayers;
+	resolvedOptions->passagePrefix =
+		GET_STRING_RELOPTION(options, passagePrefixOffset);
+	resolvedOptions->queryPrefix =
+		GET_STRING_RELOPTION(options, queryPrefixOffset);
 
-	if (useCase == PGRN_OPTION_USE_CASE_PREFIX_SEARCH)
-	{
-		*tokenizer = NULL;
-	}
-	else
-	{
-		if (PGrnIsExplicitNoneValue(rawTokenizer))
-		{
-			*tokenizer = NULL;
-		}
-		else if (PGrnIsNoneValue(rawTokenizer))
-		{
-			if (defaultTokenizerName)
-				*tokenizer = PGrnLookup(defaultTokenizerName, ERROR);
-			else
-				*tokenizer = NULL;
-		}
-		else
-		{
-			*tokenizer = &(buffers->tokenizer);
-			GRN_TEXT_SETS(ctx, *tokenizer, rawTokenizer);
-		}
-	}
+	PGrnResolveOptionValuesTokenizer(
+		options, index, i, useCase, defaultTokenizer, resolvedOptions);
 
-	PGrnApplyOptionValuesNormalizers(options,
-									 index,
-									 i,
-									 useCase,
-									 normalizers,
-									 defaultNormalizers);
+	PGrnResolveOptionValuesNormalizers(
+		options, index, i, useCase, defaultNormalizers, resolvedOptions);
 
 	rawTokenFilters = ((const char *) options) + options->tokenFiltersOffset;
 	if (!PGrnIsNoneValue(rawTokenFilters))
 	{
-		*tokenFilters = &(buffers->tokenFilters);
-		GRN_TEXT_SETS(ctx, *tokenFilters, rawTokenFilters);
+		resolvedOptions->tokenFilters = &(buffers->tokenFilters);
+		GRN_TEXT_SETS(ctx, resolvedOptions->tokenFilters, rawTokenFilters);
+	}
+
+	const char *rawPlugins = ((const char *) options) + options->pluginsOffset;
+	if (!PGrnIsNoneValue(rawPlugins))
+	{
+		resolvedOptions->plugins = &(buffers->plugins);
+		GRN_BULK_REWIND(resolvedOptions->plugins);
+		PGrnOptionParseNames(
+			rawPlugins, PGrnOptionAppendName, resolvedOptions->plugins);
 	}
 
 	lexiconTypeName = GET_STRING_RELOPTION(options, lexiconTypeOffset);
@@ -736,23 +1001,31 @@ PGrnApplyOptionValues(Relation index,
 	{
 		if (strcmp(lexiconTypeName, PGRN_LEXICON_TYPE_HASH_TABLE) == 0)
 		{
-			*lexiconType |= GRN_OBJ_TABLE_HASH_KEY;
+			resolvedOptions->lexiconType |= GRN_OBJ_TABLE_HASH_KEY;
 		}
 		else if (strcmp(lexiconTypeName, PGRN_LEXICON_TYPE_PATRICIA_TRIE) == 0)
 		{
-			*lexiconType |= GRN_OBJ_TABLE_PAT_KEY;
+			resolvedOptions->lexiconType |= GRN_OBJ_TABLE_PAT_KEY;
 		}
-		else if (strcmp(lexiconTypeName, PGRN_LEXICON_TYPE_DOUBLE_ARRAY_TRIE) == 0)
+		else if (strcmp(lexiconTypeName, PGRN_LEXICON_TYPE_DOUBLE_ARRAY_TRIE) ==
+				 0)
 		{
-			*lexiconType |= GRN_OBJ_TABLE_DAT_KEY;
+			resolvedOptions->lexiconType |= GRN_OBJ_TABLE_DAT_KEY;
 		}
 	}
 	else
 	{
-		*lexiconType |= GRN_OBJ_TABLE_PAT_KEY;
+		if (useCase == PGRN_OPTION_USE_CASE_SEMANTIC_SEARCH)
+		{
+			resolvedOptions->lexiconType |= GRN_OBJ_TABLE_HASH_KEY;
+		}
+		else
+		{
+			resolvedOptions->lexiconType |= GRN_OBJ_TABLE_PAT_KEY;
+		}
 	}
 
-	PGrnApplyOptionValuesIndexFlags(options, index, i, indexFlags);
+	PGrnResolveOptionValuesIndexFlags(options, index, i, resolvedOptions);
 }
 
 grn_expr_flags
@@ -772,80 +1045,63 @@ PGrnOptionsGetExprParseFlags(Relation index)
 }
 
 bytea *
-pgroonga_options_raw(Datum reloptions,
-					 bool validate)
+pgroonga_options(Datum reloptions, bool validate)
 {
 	PGrnOptions *grnOptions;
 	const relopt_parse_elt optionsMap[] = {
-		{"tokenizer", RELOPT_TYPE_STRING,
+		{"tokenizer",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, tokenizerOffset)},
-		{"normalizer", RELOPT_TYPE_STRING,
+		{"tokenizer_mapping",
+		 RELOPT_TYPE_STRING,
+		 offsetof(PGrnOptions, tokenizerMappingOffset)},
+		{"normalizer",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, normalizerOffset)},
-		{"token_filters", RELOPT_TYPE_STRING,
+		{"token_filters",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, tokenFiltersOffset)},
-		{"plugins", RELOPT_TYPE_STRING,
-		 offsetof(PGrnOptions, pluginsOffset)},
-		{"full_text_search_normalizer", RELOPT_TYPE_STRING,
+		{"plugins", RELOPT_TYPE_STRING, offsetof(PGrnOptions, pluginsOffset)},
+		{"full_text_search_normalizer",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, fullTextSearchNormalizerOffset)},
-		{"regexp_search_normalizer", RELOPT_TYPE_STRING,
+		{"regexp_search_normalizer",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, regexpSearchNormalizerOffset)},
-		{"prefix_search_normalizer", RELOPT_TYPE_STRING,
+		{"prefix_search_normalizer",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, prefixSearchNormalizerOffset)},
-		{"lexicon_type", RELOPT_TYPE_STRING,
+		{"lexicon_type",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, lexiconTypeOffset)},
-		{"query_allow_column", RELOPT_TYPE_BOOL,
+		{"query_allow_column",
+		 RELOPT_TYPE_BOOL,
 		 offsetof(PGrnOptions, queryAllowColumn)},
-		{"normalizers", RELOPT_TYPE_STRING,
+		{"normalizers",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, normalizersOffset)},
-		{"normalizers_mapping", RELOPT_TYPE_STRING,
+		{"normalizers_mapping",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, normalizersMappingOffset)},
-		{"index_flags_mapping", RELOPT_TYPE_STRING,
+		{"index_flags_mapping",
+		 RELOPT_TYPE_STRING,
 		 offsetof(PGrnOptions, indexFlagsMappingOffset)},
+		{"model", RELOPT_TYPE_STRING, offsetof(PGrnOptions, modelOffset)},
+		{"n_gpu_layers", RELOPT_TYPE_INT, offsetof(PGrnOptions, nGPULayers)},
+		{"passage_prefix",
+		 RELOPT_TYPE_STRING,
+		 offsetof(PGrnOptions, passagePrefixOffset)},
+		{"query_prefix",
+		 RELOPT_TYPE_STRING,
+		 offsetof(PGrnOptions, queryPrefixOffset)},
 	};
 
-#ifdef PGRN_HAVE_BUILD_RELOPTIONS
 	grnOptions = build_reloptions(reloptions,
 								  validate,
 								  PGrnReloptionKind,
 								  sizeof(PGrnOptions),
 								  optionsMap,
 								  lengthof(optionsMap));
-#else
-	{
-		relopt_value *options;
-		int nOptions;
-		options = parseRelOptions(reloptions,
-								  validate,
-								  PGrnReloptionKind,
-								  &nOptions);
-		grnOptions = allocateReloptStruct(sizeof(PGrnOptions),
-										  options,
-										  nOptions);
-		fillRelOptions(grnOptions,
-					   sizeof(PGrnOptions),
-					   options,
-					   nOptions,
-					   validate,
-					   optionsMap,
-					   lengthof(optionsMap));
-		pfree(options);
-	}
-#endif
 
 	return (bytea *) grnOptions;
 }
-
-/**
- * pgroonga.options() -- amoptions
- */
-Datum
-pgroonga_options(PG_FUNCTION_ARGS)
-{
-	Datum reloptions = PG_GETARG_DATUM(0);
-	bool validate = PG_GETARG_BOOL(1);
-	bytea *grnOptions;
-
-	grnOptions = pgroonga_options_raw(reloptions, validate);
-
-	PG_RETURN_BYTEA_P(grnOptions);
-}
